from __future__ import annotations
import logging
from typing import Any, TypeVar, Optional, List, Union, Dict
from pathlib import Path
from renku.domain_model.dataset import Dataset as RenkuDataSet
import omnibenchmark.management.general_checks
from omnibenchmark.management.data_checks import dataset_name_exist, renku_dataset_exist
from renku.command.dataset import (
    create_dataset_command,
    import_dataset_command,
    update_datasets_command,
    add_to_dataset_command,
    file_unlink_command,
    remove_dataset_command,
)
from renku.core import errors
from omnibenchmark.utils.exceptions import ProjectError

PathLike = TypeVar("PathLike", str, Path, None)
logger = logging.getLogger("omnibenchmark.renku_commands")


def renku_dataset_create(
    name: str,
    kg_url: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    creators: Optional[List[str]] = None,
    meta_data: Optional[Dict[str, Any]] = None,
    keyword: Optional[List[str]] = None,
) -> RenkuDataSet:
    """Generate an empty renku dataset in the current project. Works in a renku project only.

    Args:
        name ([str]): Dataset name
        title (str, optional): Dataset title. Defaults to None.
        description (str, optional): Description of the dataset. Defaults to None.
        creators (str, optional): Dataset creator. Defaults to None,
                                  which will be the github account activated for that renku project.
        metadata (PathLike, optional): Path to json or yaml file containing meta data information. Defaults to None.
        keyword (list[str], optional): List of keywords to be associated with the dataset. Defaults to None.

    Returns:
        RenkuDataSet: An object of class Dataset from renku.core.models.dataset
    """
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was created."
        )

    if renku_dataset_exist(name):
        print(f"Dataset {name} already exists in this repository.")
        return

    if dataset_name_exist(name, kg_url):
        print(f"Dataset {name} already taken. Please use a different name.")
        return

    result = (
        create_dataset_command()
        .build()
        .execute(
            name=name,
            title=title,
            description=description,
            creators=creators,
            keywords=keyword,
            custom_metadata=meta_data,
        )
    )
    print(
        f"Renku dataset with name {name} and the following keywords {keyword} was generated."
    )
    return result.output


def renku_dataset_import(
    uri: str, name: Optional[str] = None, extract: bool = False, yes: bool = True
):
    """Import renku dataset by url

    Args:
        uri (str): URL to the dataset to import
        name (str, optional): Name of the imported dataset. Defaults to None.
        extract (bool, optional): If the dataset is zipped and shall be extracted. Defaults to False.
        yes (bool, optional): Skip manual confirmation. Defaults to True.

    Raises:
        ProjectError: Project to import the dataset into needs to be a renku project
    """
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was imported."
        )

    if name is not None:
        if renku_dataset_exist(name):
            logger.info(f"Dataset {name} already exists in this repository.")
            return

    result = (
        import_dataset_command()
        .build()
        .execute(uri=uri, name=name, extract=extract, yes=yes)
    )

    return result.output


def renku_dataset_update(
    names: List[str],
    creators: Optional[List[str]] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    ref: Optional[str] = None,
    delete: bool = True,
    no_external: bool = True,
    no_local: bool = False,
    no_remote: bool = False,
    check_data_directory: bool = False,
    update_all: bool = False,
    dry_run: bool = False,
    plain: bool = False,
):
    """Update an imported renku dataset from source

    Args:
        names (List[str]):name of the dataset
        creators (Optional[List[str]], optional): Original dataset creator. Defaults to None.
        include (Optional[List[str]], optional): Files to include into the update. Defaults to None.
        exclude (Optional[List[str]], optional): Files to exclude from updating. Defaults to None.
        ref (Optional[str], optional): commit hash to update the dataset from. Defaults to None.
        delete (bool, optional): Delete files that do not exist in source. Defaults to True.
        no_external (bool, optional): Only update internal datasets. Defaults to True.
        update_all (bool, optional): Update all dataset files. Defaults to False.
        dry_run (bool, optional): Dry run the update. Defaults to False.

    Raises:
        ProjectError: The project needs to be a renku project
        errors.ParameterError: Specify only one of exclude/include and update_all
    """
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No dataset was updated."
        )

    if (
        not update_all
        and len(names) == 0
        and include is None
        and exclude is None
        and not dry_run
    ):
        raise errors.ParameterError(
            "Either names, update_all, dry_run, or include/exclude should be specified"
        )

    if len(names) >= 1 and update_all:
        raise errors.ParameterError("Cannot pass dataset names with update_all")
    elif (include or exclude is not None) and update_all:
        raise errors.ParameterError("Cannot pass include/exclude with update_all")

    result = (
        update_datasets_command(dry_run=dry_run)
        .build()
        .execute(
            names=names,
            creators=creators,
            include=include,
            exclude=exclude,
            ref=ref,
            delete=delete,
            no_external=no_external,
            update_all=update_all,
            dry_run=dry_run,
            no_local=no_local,
            no_remote=no_remote,
            check_data_directory=check_data_directory,
            plain=plain,
        )
    )

    return result.output


def renku_add_to_dataset(
    urls: List[str],
    dataset_name: str,
    force: bool = False,
    overwrite: bool = True,
    create: bool = False,
    datadir: Optional[Union[str, Path]] = None,
    destination: Optional[List[str]] = None,
) -> Optional[RenkuDataSet]:
    """Add files to renku dataset

    Args:
        urls (List[str]): Urls to add
        dataset_name (str): Name of the dataset
        external (bool, optional): Do urls point to external files. Defaults to False.
        force (bool, optional): Force overwriting of files. Defaults to False.
        overwrite (bool, optional): Ovrwrite files. Defaults to True.
        create (bool, optional): Create the dataset if it does not exist already. Defaults to False.
        sources (Optional[List[Union[str, Path]]], optional): Sources to add files from. Defaults to None.
        destination (Optional[List[str]], optional): Destinations of the added files. Defaults to None.

    Raises:
        ProjectError: The project needs to be a renku project
    """

    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No files were added."
        )

    if datadir is None:
        datadir = "data/" + dataset_name

    result = (
        add_to_dataset_command()
        .build()
        .execute(
            urls=urls,
            dataset_name=dataset_name,
            force=force,
            overwrite=overwrite,
            create=create,
            destination=destination,
            datadir=datadir,
        )
    )
    return result


def renku_unlink_from_dataset(
    name: str,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    yes: bool = True,
):
    """Unlink files from renku dataset

    Args:
        name (str): Dataset name to unlink files from.
        include (Optional[str], optional): Pattern of files to include. Defaults to None.
        exclude (Optional[str], optional): Pattern of files to exclude. Defaults to None.
        yes (bool, optional): confirm file unlinking. Defaults to True.

    Raises:
        ProjectError: Project needs to be a renku project to perform renku actions.
        errors.ParameterError: At least one of include or exclude need to be defined.
    """

    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No files were added."
        )
    if not include and not exclude:
        raise errors.ParameterError(
            (
                "Include or exclude filters not found.\n"
                "Please specify at least one of include and/or exclude'\n"
            )
        )

    file_unlink_command().build().execute(
        name=name, include=include, exclude=exclude, yes=yes
    )


def renku_dataset_remove(data_name: str):
    """Remove renku dataset from the current project

    Args:
        data_name (str): Name of the dataset to remove from the project

    Raises:
        ProjectError: raised if the project is not a renku project.
    """
    if not omnibenchmark.management.general_checks.is_renku_project():
        raise ProjectError(
            "Directory is not a renku project.\n"
            "Make sure you are in the correct context.\n"
            "No files were added."
        )

    remove_dataset_command().build().execute(data_name)



from omnibenchmark.management import data_checks


### Test renku_dataset_exist
def test_renku_dataset_exist_for_not_existing():
    assert not data_checks.renku_dataset_exist("quatsch")


def test_renku_dataset_exists_for_existing(get_renkuDataset_List):
    assert data_checks.renku_dataset_exist("mock_dataset")


### Test dataset_name_exist
def test_dataset_name_exist_for_not_existing(mock_dataset_query):
    assert not data_checks.dataset_name_exist("quatsch", kg_url="mock_url")


def test_dataset_name_exist_for_existing(mock_dataset_query):
    assert data_checks.dataset_name_exist("mock_dataset", kg_url="mock_url")



from typing import Mapping, Optional, List, Union, Callable, Any
from string import Template
import itertools
import json

from omnibenchmark.core.input_classes import OmniInput, OmniParameter, OutMapping
from omnibenchmark.utils.decorators import option_list, option_str
from omnibenchmark.utils.user_input_checks import empty_object_to_none
from omnibenchmark.management.parameter_checks import dict_values_to_str


def join_parameter(
    parameter: Optional[Mapping[str, str]] = None, sort_keys: bool = True
) -> str:
    """Get a joined string with all parameter names and their value

    Args:
        parameter (Optional[Mapping[str, str]], optional): Parameter mapping. Defaults to None.

    Returns:
        str: Joined string
    """
    if parameter is None:
        return ""
    param_k = (
        sorted(parameter.keys(), key=lambda x: x.lower())
        if sort_keys
        else parameter.keys()
    )
    all_params = "__".join(
        ["{}_{}".format(str(par_k), parameter[par_k]) for par_k in param_k]
    )
    return all_params


@option_str
def join_inputs_parameter(
    input: Optional[str] = None,
    parameter: Optional[Mapping[str, str]] = None,
    sort_keys: bool = True,
) -> str:
    """Get a joined string of parameter and inputs with their respective values

    Args:
        input (Optional[str], optional): Input types and values. Defaults to None.
        parameter (Optional[Mapping[str, str]], optional): Parameter names and values. Defaults to None.

    Returns:
        str: A joined string
    """
    param_str = join_parameter(parameter, sort_keys=sort_keys)
    try:
        if param_str == "":
            joined = "".join([input, param_str])  # type: ignore
        else:
            joined = "__".join([input, param_str])  # type: ignore
    except Exception:
        joined = param_str
    finally:
        return joined


def get_out_names_from_input_params(
    output_end: Mapping[str, str],
    name: str,
    input: Optional[str] = None,
    parameter: Optional[Mapping[str, str]] = None,
    sort_keys: bool = True,
    out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
    **kwargs,
) -> Mapping[str, str]:
    """Generates automatic output file names from inputs and parameters used, based on a template defined in out_template.

    Args:
        input (str): Specific input for this output. Should be potentially uniquely identifyable shortcuts, e.g, dataset names.
        output_end (Mapping[str, str]): Mapping of all output file types and their corresponding endings,
                                        e.g. {'norm_counts': 'mtx.gz'}.
        name (str): Name of the method, dataset, etc. the output belongs to.
        parameters (Optional[Mapping[str, str]], optional): Mapping of parameter and the values used to generate these outputs.
                                                            Defaults to None.
        out_template (str, optional): Template to generate names from.

    Returns:
        Mapping[str, str]: A mapping of output filetypes and generated names,
                           e.g. '{norm_counts: "data/meth1/meth1__data1__norm_counts.mtx.gz"}'.
    """

    # check template
    valid_keys = ["unique_values", "name", "out_name", "out_end"]
    for arg in kwargs.keys():
        valid_keys.append(arg)
    temp_keys = [
        s[1] or s[2] for s in Template.pattern.findall(out_template) if s[1] or s[2]
    ]
    if any(temp not in valid_keys for temp in temp_keys):
        raise NameError(
            f"Invalid output filename template: {out_template}."
            f"Please use valid keys as in {valid_keys} or consider specifying outputs explicitly."
        )
    unique_values = join_inputs_parameter(input, parameter, sort_keys=sort_keys)
    sub_dict = kwargs
    sub_dict["unique_values"] = unique_values
    sub_dict["name"] = name
    output_names = {}
    for out in output_end.keys():
        out_end = output_end[out]
        sub_dict["out_end"] = out_end
        sub_dict["out_name"] = out
        output_file = Template(out_template).substitute(**sub_dict)
        output_names[out] = output_file.replace("..", ".")

    return output_names


@option_list
def get_input_file_list(inputs: Optional[OmniInput]) -> List:
    """Get a List with all input file types

    Args:
        inputs (Optional[OmniInput]): An OmniInput object

    Returns:
        List: A list with all input file types
    """
    return (
        list(inputs.input_files.keys())  # type: ignore
        if inputs.input_files is not None  # type: ignore
        else []
    )


@option_list
def get_parameter_combinations(parameter: Optional[OmniParameter]) -> List:
    """Extract parameter combinations as specified in an OmniParameter object

    Args:
        parameter (Optional[OmniParameter]): An OmniParameter object

    Returns:
        List: List with all specified parameter combinations
    """
    return (
        parameter.combinations  # type: ignore
        if parameter.combinations is not None  # type: ignore
        else []
    )


@option_list
def get_input_parameter_combinations(
    inputs: Optional[OmniInput] = None, parameter: Optional[OmniParameter] = None
) -> List:
    """Get a list with all parameter and input combinations as OutFileMapping, but no outputs, yet

    Args:
        inputs (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        parameter (Optional[OmniParameter], optional): An OmniParameter object. Defaults to None.

    Returns:
        List: _description_
    """
    ins = get_input_file_list(inputs)
    para = get_parameter_combinations(parameter)
    out_list = []

    all_comb = list(itertools.product(ins, para))
    if len(all_comb) >= 1:
        for comb in all_comb:
            out_dict: OutMapping = {
                "output_files": None,
                "input_files": {comb[0]: inputs.input_files[comb[0]]},  # type: ignore
                "parameter": comb[1],
            }
            out_list.append(out_dict)
    else:
        for in_name in ins:
            out_dict_in: OutMapping = {
                "output_files": None,
                "input_files": {in_name: inputs.input_files[in_name]},  # type: ignore
                "parameter": None,
            }
            out_list.append(out_dict_in)
        for para_dict in para:
            out_dict_para: OutMapping = {
                "output_files": None,
                "input_files": None,
                "parameter": para_dict,
            }
            out_list.append(out_dict_para)

    return out_list


def get_template_vars_dependencies(
    comb: OutMapping, fun: Callable[..., Mapping], **kwargs
) -> Mapping:
    return fun(comb, **kwargs)


def get_all_output_combinations(
    name: str,
    output_end: Mapping[str, str],
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
    sort_keys: bool = True,
    out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
    template_fun: Optional[Callable[..., Mapping]] = None,
    **kwargs,
) -> List[OutMapping]:
    """Get OutMappings for all possible input and parameter combinations

    Args:
        name (str): Output name
        output_end (Mapping[str, str]): Mapping of output file types with their corresponding file endings
        inputs (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        parameter (Optional[OmniParameter], optional): An OmniParameter object. Defaults to None.
        out_template (str, optional): A template to automatically generate output file names. Defaults to "data//__.".
        template_fun (Optional[Callable[..., Mapping]], optional): Function to apply for automatic output file generation.
                                                                   Defaults to None.

    Returns:
        List[OutMapping]: A List of all possible OutMappings
    """
    comb_list = get_input_parameter_combinations(inputs=inputs, parameter=parameter)
    out_list = []

    for comb in comb_list:
        if template_fun is not None:
            temp_vars = get_template_vars_dependencies(
                comb=comb, fun=template_fun, **kwargs
            )
        else:
            temp_vars = kwargs
        try:
            input_str = next(iter(comb["input_files"].keys()))
            comb["input_files"] = comb["input_files"][input_str]
        except Exception:
            input_str = None
        finally:
            out_names = get_out_names_from_input_params(
                name=name,
                output_end=output_end,
                input=input_str,
                parameter=comb["parameter"],
                sort_keys=sort_keys,
                out_template=out_template,
                **temp_vars,
            )
        comb["output_files"] = out_names
        out_list.append(comb)

    if len(out_list) == 0:
        out_names = get_out_names_from_input_params(
            name=name,
            output_end=output_end,
            sort_keys=sort_keys,
            out_template=out_template,
        )
        out_map: OutMapping = {
            "output_files": out_names,
            "input_files": None,
            "parameter": None,
        }
        out_list.append(out_map)
    return out_list


@option_list
def convert_values_to_string(d: Optional[Mapping]):
    return {key: str(d[key]) for key in d.keys()}  # type: ignore


@option_list
def get_default(omni_class: Optional[Union[OmniParameter, OmniInput]]):
    default_val = omni_class.default  # type: ignore
    if isinstance(default_val, dict):
        default_val = convert_values_to_string(default_val)
    return default_val


@option_list
def get_default_input(omni_input: Optional[OmniInput]):
    def_input_nam = omni_input.default  # type: ignore
    if def_input_nam is not None:
        return convert_values_to_string(
            omni_input.input_files[def_input_nam]  # type: ignore
        )
    else:
        return []


def check_default_settings(
    default_inputs: Optional[Mapping],
    default_outputs: Optional[Mapping],
    default_params: Optional[Mapping],
    defaults: OutMapping,
) -> OutMapping:
    """Check if the specified defaults for inputs/parameter/outputs match the specified default OutMapping

    Args:
        default_inputs (Optional[Mapping]): Default inputs
        default_outputs (Optional[Mapping]): Default outputs
        default_params (Optional[Mapping]): Default parameter
        defaults (OutMapping): Default OutMapping

    Raises:
        InputError: Specified defaults need to match.

    Returns:
        OutMapping: Default OutMapping
    """
    if (
        default_outputs == defaults["output_files"]
        and default_params == convert_values_to_string(defaults["parameter"])
        and default_inputs == convert_values_to_string(defaults["input_files"])
    ):
        return defaults
    else:
        print(
            f"WARNING: Default mappings do not match.\n"
            f"If you did not specify defaults, ignore this warning.\n"
            f"Otherwise, please check that your combination is valid.\n"
            f"The following default file mapping is used:\n {defaults}"
        )
        return defaults


def get_default_outputs(
    file_mapping: List[OutMapping],
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
) -> Optional[Mapping]:
    """Get default outputs from the file mapping and default inputs/parameter

    Args:
        file_mapping (List[OutMapping]): List of OutMappings
        inputs (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        parameter (Optional[OmniParameter], optional): An OmniParameter object. Defaults to None.

    Returns:
        Optional[Mapping]: Default Output files
    """
    def_input = get_default_input(inputs)
    def_param = get_default(parameter)
    def_out = None
    for out_dict in file_mapping:
        if (
            not all(
                element is None
                for element in [out_dict["input_files"], out_dict["parameter"]]
            )
            and convert_values_to_string(out_dict["input_files"]) == def_input
            and convert_values_to_string(out_dict["parameter"]) == def_param
        ):
            def_out = out_dict["output_files"]
            def_map = out_dict
    if isinstance(def_out, type(None)):
        def_out = file_mapping[0]["output_files"]
        def_map = file_mapping[0]
    def_map = check_default_settings(
        default_inputs=def_input,
        default_outputs=def_out,
        default_params=def_param,
        defaults=def_map,
    )
    return def_out


def add_missing_file_map_keys(file_map: OutMapping) -> OutMapping:
    """Add 'input_files', 'parameter' as keys to OutMapping if they are missing

    Args:
        file_map (OutMapping): OutMapping

    Returns:
        OutMapping: OutMapping with all keys.
    """
    for map_key in ["input_files", "parameter"]:
        if map_key not in file_map.keys():
            file_map[map_key] = None  # type:ignore
    return file_map


def autocomplete_file_mapping(file_mapping: List[OutMapping]) -> List[OutMapping]:
    return [add_missing_file_map_keys(file_map) for file_map in file_mapping]


@option_list
def get_input_param_names(
    in_obj: Optional[Union[OmniInput, OmniParameter]]
) -> List[str]:
    return in_obj.names  # type:ignore


@option_list
def get_keys_list(map: Mapping[str, Any]) -> List[str]:
    return list(map.keys())  # type:ignore


def filter_file_mapping_missing_values(
    file_mapping: OutMapping,
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
) -> Optional[OutMapping]:
    """Filter incomplete file mappings (e.g., if a file type specified as input is missing)

    Args:
        file_mapping (OutMapping): An OutMapping
        inputs (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        parameter (Optional[OmniParameter], optional): An OmniParameter object. Defaults to None.

    Returns:
        Optional[OutMapping]: Complete OutMapping
    """
    in_files = get_input_param_names(inputs)
    in_param = get_input_param_names(parameter)
    if any(fi not in get_keys_list(file_mapping["input_files"]) for fi in in_files):
        return None
    elif any(
        param not in get_keys_list(file_mapping["parameter"]) for param in in_param
    ):
        return None
    else:
        return file_mapping


def rm_key(d: Mapping, key: str) -> Mapping:
    new_d = dict(d)
    del new_d[key]
    return new_d


def filter_file_mapping_list_input_param_combinations(
    file_mapping: List[OutMapping], filter_list: Mapping
) -> List[OutMapping]:
    """Filter specific input, parameter combinations from the output file mapping list.

    Args:
        file_mapping (List[OutMapping]): List of Output file mappings
        filter_list (Optional[Mapping], optional): List with combinations to filter.

    Returns:
        List[OutMapping]: Filtered list of OutMappings
    """
    filter_format = [
        {filt_key: dict_values_to_str(filt_val) for filt_key, filt_val in filt.items()}
        for filt in filter_list
    ]

    return [
        comb
        for comb in file_mapping
        if rm_key(
            {
                file_key: dict_values_to_str(file_val)  # type:ignore
                for file_key, file_val in comb.items()
            },
            "output_files",
        )
        not in filter_format
    ]


def filter_file_mapping_list_input_param_combinations_json(
    file_mapping: List[OutMapping], filter_json: Optional[str] = None
) -> List[OutMapping]:
    """Filter specific input, parameter combinations defined in a filter.json file from the output file mapping list.

    Args:
        file_mapping (List[OutMapping]): Mapping between input, output and parameter files
        filter_json (Optional[str], optional): Path to json file with the specified combinations. Defaults to None.

    Returns:
        List[OutMapping]: Filtered list of OutMappings
    """
    try:
        with open(filter_json) as f:  # type:ignore
            filter_list = json.load(f)
    except (FileNotFoundError, TypeError):
        return file_mapping

    return filter_file_mapping_list_input_param_combinations(
        file_mapping=file_mapping, filter_list=filter_list
    )


def filter_file_mapping_list(
    file_mapping_list: List[OutMapping],
    inputs: Optional[OmniInput] = None,
    parameter: Optional[OmniParameter] = None,
    filter_json: Optional[str] = None,
) -> Optional[List[OutMapping]]:
    """Filter list of file mappings for imcomplete mappings

    Args:
        file_mapping_list (List[OutMapping]): List of OutMappings
        inputs (Optional[OmniInput], optional): An OmniInput object. Defaults to None.
        parameter (Optional[OmniParameter], optional): An OmniParameter object. Defaults to None.

    Returns:
        Optional[List[OutMapping]]: List of complete OutMappings
    """
    fi_map_list = [
        filter_file_mapping_missing_values(
            file_mapping=fi_map, inputs=inputs, parameter=parameter
        )
        for fi_map in file_mapping_list
    ]
    res = list(filter(None, fi_map_list))
    res_comb = filter_file_mapping_list_input_param_combinations_json(
        file_mapping=res, filter_json=filter_json
    )
    res_uni = [i for n, i in enumerate(res_comb) if i not in res_comb[n + 1 :]]
    return empty_object_to_none(res_uni)



"""Fixtures related to utils functions"""

import pytest
from renku.domain_model.dataset import Dataset as RenkuDataset
from renku.domain_model.dataset import DatasetFile as RenkuDatasetFile
from renku.domain_model.entity import Entity
from renku.ui.api.models.dataset import Dataset as ApiDataset
import yaml


@pytest.fixture
def mock_param():
    return {"param_str": "value_str", "param_num": 10}


@pytest.fixture
def mock_out_end():
    return {"red_dim": "mtx.gz", "counts": "json"}


@pytest.fixture
def mock_config():
    with open("omnibenchmark/test/utils/ex_config.yaml") as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config


@pytest.fixture
def mock_paramDatasetFile():
    ent = Entity(path="omnibenchmark/test/utils/ex_param.json", checksum="XDFGHJK")
    mock_data_fi = RenkuDatasetFile(entity=ent)
    return mock_data_fi


@pytest.fixture
def mock_paramDataset(mock_paramDatasetFile):
    mock_data = RenkuDataset(name="param_dataset")
    mock_data.keywords = ["mock_param"]
    mock_data.title = "Mock param dataset"
    mock_data.dataset_files = [mock_paramDatasetFile, mock_paramDatasetFile]
    return mock_data


@pytest.fixture
def mock_paramapi_Dataset(mock_paramDataset):
    api_data = ApiDataset()
    api_data = api_data._from_dataset(mock_paramDataset)
    return api_data



"""Renku workflow commands"""

from __future__ import annotations
from typing import List, Optional
from omnibenchmark.core.input_classes import PathLike
from renku.command.run import run_command
from renku.command.workflow import execute_workflow_command, revert_activity_command
from renku.command.update import update_command
from renku.command.command_builder.command import CommandResult
from renku.core import errors
from renku.core.util.metadata import construct_creators
from renku.command.format.activity import tabulate_activities
from omnibenchmark.management.general_checks import is_renku_project
import logging

logger = logging.getLogger("omnibenchmark.renku_commands")


def renku_workflow_run(
    command_line: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    keyword: Optional[str] = None,
    success_codes: Optional[List[int]] = None,
    explicit_inputs: List = [],
    explicit_outputs: List = [],
    explicit_parameters: List = [],
    no_output: bool = False,
    no_input_detection: bool = False,
    no_output_detection: bool = False,
    creators: Optional[List[str]] = None,
) -> CommandResult:
    """Run a renku workflow

    Args:
        command_line (str): Command to run
        name (str, optional): Workflow name. Defaults to None.
        description (str, optional): Workflow description. Defaults to None.
        keyword (str, optional): Workflow keyword. Defaults to None.
        success_codes (List[int], optional): List of accepted auccess codes. Defaults to None.
        explicit_inputs (List, optional): Explicit input file paths. Defaults to [].
        explicit_outputs (List, optional): Explicit output file paths. Defaults to [].
        explicit_parameters (List, optional): Explicit parameter. Defaults to [].
        no_output (bool, optional): Workflow does not generate outputs. Defaults to False.
        no_input_detection (bool, optional): Turn automatic input detection off. Defaults to False.
        no_output_detection (bool, optional): Turn automatic output detection off. Defaults to False.
    """
    if not is_renku_project():
        logger.error(
            "Project is not a renku project. No workflow generated, consider to running your command without renku."
        )
        return

    if creators:
        creators, _ = construct_creators(creators)

    workflow = (
        run_command()
        .build()
        .execute(
            name=name,
            description=description,
            keyword=keyword,
            explicit_inputs=explicit_inputs,
            explicit_outputs=explicit_outputs,
            explicit_parameters=explicit_parameters,
            no_output=no_output,
            no_input_detection=no_input_detection,
            no_output_detection=no_output_detection,
            success_codes=success_codes,
            command_line=command_line,
            creators=creators,
        )
    )

    return workflow


def renku_workflow_execute(
    name_or_id: str,
    set_params: List[str],
    provider: str = "cwltool",
    config: Optional[str] = None,
    values: Optional[str] = None,
    skip_metadata_update: bool = False,
):
    """Execute an existing renku workflow with a specific input/parameter/output combination

    Args:
        name_or_id (str): Workflow name or id
        set_params (List[str]): List of the parameter/inputs/outputs values to use
        provider (str, optional): Provider to run the workflow. Defaults to "cwltool".
        config (Optional[str], optional):Config file for the provider. Defaults to None.
        values (Optional[str], optional): Values for the provider to use. Defaults to None.
        skip_metadata_update (bool, optional): Do not update the workflow metadata. Defaults to False.

    Returns:
        _type_: _description_
    """
    workflow = (
        execute_workflow_command(skip_metadata_update=skip_metadata_update)
        .build()
        .execute(
            name_or_id=name_or_id,
            set_params=set_params,
            provider=provider,
            config=config,
            values=values,
        )
    )
    return workflow


def renku_update_activity(
    update_all: bool = False,
    paths: Optional[List[PathLike]] = None,
    dry_run: bool = False,
    provider: str = "cwltool",
    ignore_deleted: bool = True,
    config: Optional[str] = None,
    skip_metadata_update: bool = False,
):
    """Update renku existing activities

    Args:
        update_all (bool, optional): Should all activities be updated. Defaults to False.
        paths (Optional[List[PathLike]], optional): Output paths to update the activities for. Defaults to None.
        dry_run (bool, optional): Show what activities would be updated. Defaults to False.
        provider (str, optional):Provider to run the updates. Defaults to "cwltool".
        ignore_deleted (bool, optional): Ignore deleted activities. Defaults to True.
        config (Optional[str], optional): Config file for the provider. Defaults to None.
        skip_metadata_update (bool, optional): Do not update the workflow metadata. Defaults to False.
    """
    try:
        result = (
            update_command(skip_metadata_update=skip_metadata_update)
            .build()
            .execute(
                update_all=update_all,
                dry_run=dry_run,
                paths=paths,
                provider=provider,
                config=config,
                ignore_deleted=ignore_deleted,
            )
        )
    except errors.NothingToExecuteError:
        print(f"Nothing to update/execute for {paths}.")
    else:
        if dry_run:
            activities, modified_inputs = result.output
            print(tabulate_activities(activities, modified_inputs))


def renku_workflow_revert(
    activity_id: str,
    metadata_only: bool = False,
    force: bool = False,
    plan: bool = False,
):
    """Revert an existing activity by id and remove the generated outputs and corresponding plan

    Args:
        activity_id (str): activity id to be removed.
        metadata_only (bool, optional): remove metadata only, but not the generated outputs. Defaults to False.
        force (bool, optional): force, even when activity is part of a composed workflow. Defaults to False.
        plan (bool, optional): remove the corresponding plan that generated the activity as well. Defaults to False.

    Raises:
        errors.ParameterError: If reverting the activity will break downstream workflows
    """
    try:
        revert_activity_command().build().execute(
            metadata_only=metadata_only,
            force=force,
            delete_plan=plan,
            activity_id=activity_id,
        )
    except errors.ActivityDownstreamNotEmptyError:
        raise errors.ParameterError(
            "Activity has downstream dependent activities: Pass '--force' if you want to revert the activity anyways."
        )



from omnibenchmark.management import data_commands
import requests
import re

import pytest

# query_datasets_by_string
@pytest.mark.api_call
def test_query_datasets_by_string_works():
    query_res = data_commands.query_datasets_by_string("test")
    assert isinstance(query_res, list)
    assert "name" in query_res[0].keys()


# query_datasets_by_property
def test_query_datasets_by_property_with_matching_key(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res = data_commands.query_datasets_by_property("mock")
    assert res[0]["name"] == "mock_dataset"


def test_query_datasets_by_property_not_matching_key(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res = data_commands.query_datasets_by_property("something")
    assert res == []


def test_query_datasets_by_property_other_property(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res = data_commands.query_datasets_by_property(
        string="mock", match_string="XXXXXX", property_name="identifier"
    )
    assert res[0]["name"] == "mock_dataset"


# filter_existing
def test_filter_existing_with_existing(get_renkuDataset_List, mock_dataset_json):
    imp_list, up_list = data_commands.filter_existing(mock_dataset_json)
    assert imp_list == []
    assert up_list == mock_dataset_json


def test_filter_existing_not_existing(get_renkuDataset_List, mock_dataset_json):
    mock_dataset_json[0]["name"] = "something"
    imp_list, up_list = data_commands.filter_existing(mock_dataset_json)
    assert imp_list == mock_dataset_json
    assert up_list == []


@pytest.mark.api_call
def test_get_data_info_by_id_works():
    query_res = data_commands.query_datasets_by_string("test")
    query_id = query_res[0]["identifier"]
    data_info = data_commands.get_data_info_by_id(query_id)
    assert data_info["name"] == query_res[0]["name"]
    assert "url" in data_info.keys()


# get_field_by_dataset_property
def test_get_field_by_dataset_property_no_filter(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res1, res2 = data_commands.get_field_by_dataset_property("mock")
    assert res1 == ["XXXXXX"]
    assert res2 == []


def test_get_field_by_dataset_property_filter_names(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res1, res2 = data_commands.get_field_by_dataset_property(
        "mock", filter_names="mock_dataset"
    )
    assert res1 == []
    assert res2 == []


def test_get_field_by_dataset_property_with_filter(
    mock_dataset_json, monkeypatch, get_renkuDataset_List
):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res1, res2 = data_commands.get_field_by_dataset_property("mock", filter_ex=True)
    assert res1 == []
    assert res2 == ["mock_dataset"]


def test_get_field_by_dataset_property_with_filter_no_match(
    mock_dataset_json, monkeypatch, get_renkuDataset_List
):
    mock_dataset_json[0]["name"] = "something"

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res1, res2 = data_commands.get_field_by_dataset_property("mock", filter_ex=True)
    assert res1 == ["XXXXXX"]
    assert res2 == []


def test_get_field_by_dataset_property_no_data(
    mock_dataset_json, monkeypatch, get_renkuDataset_List
):
    class MockResponse:
        @staticmethod
        def json():
            return []

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    res1, res2 = data_commands.get_field_by_dataset_property("mock", filter_ex=True)
    assert res1 == []
    assert res2 == []


# find_dataset_linked_to_wflow
@pytest.mark.api_call
def test_find_dataset_linked_to_wflow(mock_dataset_info):
    linked = data_commands.find_dataset_linked_to_wflow([mock_dataset_info])
    assert linked == []


# filter_duplicated_names
@pytest.mark.api_call
def test_filter_duplicated_names_works(
    mock_dataset_info, monkeypatch, mock_dataset_json
):
    mock_copy = mock_dataset_info.copy()
    mock_copy["identifier"] = "XXXXXX"
    info_list = [mock_dataset_info, mock_copy]

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    uni_dat = data_commands.filter_duplicated_names(info_list)
    assert uni_dat == [mock_copy]


@pytest.mark.api_call
def test_filter_duplicated_names_identical_data(
    mock_dataset_info, monkeypatch, mock_dataset_json
):
    mock_copy = mock_dataset_info.copy()
    mock_copy["name"] = "new"
    info_list = [mock_dataset_info, mock_copy]
    info_list = [mock_dataset_info, mock_dataset_info]

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    uni_dat = data_commands.filter_duplicated_names(info_list)
    assert uni_dat == [mock_dataset_info]


@pytest.mark.api_call
def test_filter_duplicated_names_no_dup(
    mock_dataset_info, monkeypatch, mock_dataset_json
):
    info_list = [mock_dataset_info, mock_dataset_info]

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    uni_dat = data_commands.filter_duplicated_names(info_list)
    assert uni_dat == [mock_dataset_info]


# get_origin_dataset_ids
def test_get_origin_dataset_ids_unique_id(monkeypatch, mock_dataset_info):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_info

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    data_ids = data_commands.get_origin_dataset_ids(["idx"])
    assert data_ids == [mock_dataset_info]


def test_get_origin_dataset_ids_with_sameas(monkeypatch, mock_dataset_info):

    mock_dataset_info["sameAs"] = "idy"

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_info

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    data_ids = data_commands.get_origin_dataset_ids(["idx"])
    assert data_ids == [mock_dataset_info]


def test_get_origin_dataset_ids_with_noname(monkeypatch, mock_dataset_info):

    del mock_dataset_info["name"]

    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_info

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    data_ids = data_commands.get_origin_dataset_ids(["idx"])
    assert data_ids == []


@pytest.mark.api_call
def test_get_origin_dataset_ids_works():
    query_res = data_commands.query_datasets_by_string("test")
    query_id = query_res[0]["identifier"]

    data_ids = data_commands.get_origin_dataset_ids([query_id])
    assert data_ids[0]["name"] == query_res[0]["name"]


# Check_orchestrator
@pytest.mark.api_call
def test_check_orchestrator_existing(monkeypatch, mock_dataset_info):
    o_url = "https://renkulab.io/knowledge-graph/projects/omnibenchmark/orchestrator"
    project_info = {"identifier": 12385}
    project_mod = {"identifier": 12779}

    class MockResponse:
        @staticmethod
        def json():
            return project_info

    class MockMod:
        @staticmethod
        def json():
            return project_mod

    def get_project_info(*args, **kwargs):
        if mock_dataset_info["project"]["_links"][0]["href"] in args:
            return MockMod()
        else:
            return MockResponse()

    monkeypatch.setattr(requests, "get", get_project_info)
    o_check = data_commands.check_orchestrator(mock_dataset_info, o_url="some/url")
    assert o_check == mock_dataset_info["url"]


# get_data_url_by_keyword
def test_get_data_url_by_keyword_no_datasets(mock_dataset_json, monkeypatch):
    class MockResponse:
        @staticmethod
        def json():
            return []

    def mock_get(*args, **kwargs):
        return MockResponse()

    monkeypatch.setattr(requests, "get", mock_get)

    assert data_commands.get_data_url_by_keyword("mock", "some/path") == ([], [])


# link_files_by_prefix
def test_link_files_by_prefix_existing_file(get_renkuDataset_List, capsys):
    data_commands.link_files_by_prefix(
        keyword="mock", prefix=["genes"], data_name="some", dry_run=True
    )
    captured = capsys.readouterr()
    assert re.match(
        r"Run link_files_by_prefix with dry_run = False to link the following files to some:\n\nsome/path/to/genes_file.txt*",
        captured.out,
    )


def test_link_files_by_prefix_non_keyword(get_renkuDataset_List, capsys):
    data_commands.link_files_by_prefix(
        keyword="quatsch", prefix=["genes"], data_name="some", dry_run=True
    )
    captured = capsys.readouterr()
    assert re.match(
        r"Run link_files_by_prefix with dry_run = False to link the following files to some:\n\n\n",
        captured.out,
    )


def test_link_files_by_prefix_prefix_string(get_renkuDataset_List, capsys):
    data_commands.link_files_by_prefix(
        keyword="mock", prefix="genes", data_name="some", dry_run=True
    )
    captured = capsys.readouterr()
    assert re.match(
        r"Run link_files_by_prefix with dry_run = False to link the following files to some:\n\nsome/path/to/genes_file.txt*",
        captured.out,
    )



"""Fixtures related to management functions"""

import pytest
import json


@pytest.fixture
def mock_filter():
    return {
        "param_1": {"upper": 50, "lower": 10},
        "param_2": {"exclude": ["val1", "val3"]},
        "file": "path/to/filter.json",
    }



from omnibenchmark.core.input_classes import OmniInput, OmniParameter
from omnibenchmark.core.output_classes import OmniCommand, OmniOutput, OmniPlan
from omnibenchmark.renku_commands.datasets import (
    renku_dataset_create,
    renku_dataset_update,
)
from omnibenchmark.management.run_commands import (
    manage_renku_plan,
    manage_renku_activities,
    check_omni_command,
    get_all_output_file_names,
    check_output_directories,
    revert_run,
)
from omnibenchmark.management.general_checks import find_orchestrator, is_renku_project
from omnibenchmark.management.data_checks import find_outputs_with_missing_inputs
from omnibenchmark.management.data_commands import (
    update_dataset_files,
    get_data_url_by_keyword,
)
from omnibenchmark.utils.exceptions import InputError

from omnibenchmark.utils.user_input_checks import flatten, rm_none_from_list
from omnibenchmark.management.wflow_checks import (
    check_plan_exist,
    filter_activity_exist,
)
from omnibenchmark.utils.default_global_vars import OmniRenkuInst
from renku.domain_model.dataset import Dataset as RenkuDataSet
from renku.command.view_model.plan import PlanViewModel
from typing import List, Optional
from os import PathLike


class OmniObject(OmniRenkuInst):

    """OmniBench object

    An Omnibenchmark component. Class to store metadata, inputs, outputs and parameter of an omnibenchmark project.
    """

    def __init__(
        self,
        name: str,
        keyword: Optional[List[str]] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        script: Optional[PathLike] = None,
        command: Optional[OmniCommand] = None,
        inputs: Optional[OmniInput] = None,
        parameter: Optional[OmniParameter] = None,
        outputs: Optional[OmniOutput] = None,
        omni_plan: Optional[OmniPlan] = None,
        benchmark_name: Optional[str] = None,
        orchestrator: Optional[str] = None,
        wflow_name: Optional[str] = None,
        dataset_name: Optional[str] = None,
    ):
        """An object to manage an omnibenchmark module

        Args:
            name (str): Module name
            keyword (Optional[List[str]], optional): Keyword associated to the modules output dataset. Defaults to None.
            title (Optional[str], optional): Title of the modules output dataset. Defaults to None.
            description (Optional[str], optional): Description of the modules output dataset. Defaults to None.
            script (Optional[PathLike], optional): Script to generate the modules workflow for. Defaults to None.
            command (Optional[OmniCommand], optional): Workflow command, automatically generated if missing. Defaults to None.
            inputs (Optional[OmniInput], optional): Definitions of the workflow inputs. Defaults to None.
            parameter (Optional[OmniParameter], optional): Definitions of the workflow parameter. Defaults to None.
            outputs (Optional[OmniOutput], optional): Definitions of the workflow outputs. Defaults to None.
            omni_plan (Optional[OmniPlan], optional): The workflow description. Defaults to None.
            benchmark_name (Optional[str], optional): Name of the benchmark the module is associated to. Defaults to None.
            orchestrator (Optional[str], optional): Orchestrator url of the benchmark th emodule is associated to.
                                                    Automatic detection. Defaults to None.
            wflow_name (Optional[str], optional): Workflow name. Will be set to the module name if none. Defaults to None.
            dataset_name (Optional[str], optional): Dataset name. Will be set to the module name if none. Defaults to None.
        """
        self.name = name
        self.keyword = keyword
        self.title = title
        self.description = description
        self.command = command
        self.inputs = inputs
        self.outputs = outputs
        self.parameter = parameter
        self.script = script
        self.omni_plan = omni_plan
        self.orchestrator = orchestrator
        self.benchmark_name = benchmark_name
        self.wflow_name = wflow_name
        self.dataset_name = dataset_name
        self.renku: bool = is_renku_project()
        self.kg_url: str = super().KG_URL

        if self.command is None and self.script is not None:
            self.command = OmniCommand(script=self.script, outputs=self.outputs)

        if self.orchestrator is None and self.benchmark_name is not None:
            self.orchestrator = find_orchestrator(
                benchmark_name=self.benchmark_name, bench_url=self.BENCH_URL
            )

        if self.wflow_name is None:
            self.wflow_name = self.name

        if self.dataset_name is None:
            self.dataset_name = self.name

    def create_dataset(self) -> RenkuDataSet:
        """create renku-dataset defined by the attributes of the class instance

        Returns:
            RenkuDataSet: An object of class Dataset from renku.core.models.dataset containing the class instances attributes.
        """

        renku_dataset = renku_dataset_create(
            self.dataset_name,  # type:ignore
            self.kg_url,
            title=self.title,
            description=self.description,
            keyword=self.keyword,
        )
        return renku_dataset

    def run_renku(self, save: bool = False, all: bool = True):
        self.command = check_omni_command(self.command, self.script, self.outputs)
        if self.outputs is not None:
            out_files = get_all_output_file_names(self.outputs)
            check_output_directories(out_files)
        self.omni_plan = manage_renku_plan(
            omni_plan=self.omni_plan,
            command=self.command,
            output=self.outputs,
            name=self.wflow_name,
            description=self.description,
        )
        if self.outputs is not None and all:
            manage_renku_activities(
                outputs=self.outputs, omni_plan=self.omni_plan, save=save
            )

    def update_result_dataset(self, clean: bool = True):
        """Add output files to the objects dataset and update the dataset"""
        if self.outputs is None:
            print(
                f"No output files detected. Nothing to update for {self.dataset_name}."
            )
            return
        out_files = get_all_output_file_names(self.outputs)
        update_dataset_files(
            urls=out_files, dataset_name=self.dataset_name  # type:ignore
        )
        renku_dataset_update(names=[self.dataset_name])  # type:ignore
        out_no_input = find_outputs_with_missing_inputs()
        if len(out_no_input) > 0:
            revert_run(
                out_files=out_no_input, dataset_name=self.dataset_name, remove=clean
            )

    def update_object(
        self, check_o_url: bool = True, n_latest: int = 9, all: bool = True
    ):
        """Update the objects inputs, parameter and output definition. Does not run or update workflows/activities.
        Args:
            check_o_url (bool): If linking to an orchestrator shall be checked.
                                WARNING: If False ALL existing datasets with that keyword will be imported!
            n_latest (int): Number of latest pipelines to include into orchestrator checks
        """
        if not check_o_url:
            self.orchestrator = "placeholder/string"
        if self.orchestrator is None:
            if self.benchmark_name is not None:
                self.orchestrator = find_orchestrator(
                    benchmark_name=self.benchmark_name, bench_url=self.BENCH_URL
                )
            if self.orchestrator is None:
                print(
                    f"WARNING: No orchestrator specified! \n"
                    f"No new datasets will be imported. Consider specifying an orchestrator by running:\n"
                    f"OmniObject.orchestrator = find_orchestrator(BENCHMARK_NAME) \n"
                    f"Look at {self.BENCH_URL} to get a list of possible BENCHMARK_NAMEs."
                )
        if self.inputs is not None and self.orchestrator is not None:
            self.inputs.update_inputs(
                orchestrator=self.orchestrator,
                query_url=self.DATA_QUERY_URL,
                data_url=self.DATA_URL,
                gitlab_url=self.GIT_URL,
                check_o_url=check_o_url,
                n_latest=n_latest,
                all=all,
            )
        if self.parameter is not None and self.orchestrator is not None:
            self.parameter.update_parameter(
                orchestrator=self.orchestrator,
                query_url=self.DATA_QUERY_URL,
                data_url=self.DATA_URL,
                gitlab_url=self.GIT_URL,
                check_o_url=check_o_url,
                n_latest=n_latest,
            )
        if self.outputs is not None:
            self.outputs.inputs = self.inputs
            self.outputs.parameter = self.parameter
            self.outputs.update_outputs()

        if self.command is not None:
            self.command.outputs = self.outputs
            self.command.update_command()

    def clean_revert_run(self):
        """Unlink and delete all output files, and revert all activities related to this object and delete the corresponding plan.

        This is intended in cases of overthroughing everything and starting with a clean repository.
        """
        out_files = flatten(
            [list(fi["output_files"].values()) for fi in self.outputs.file_mapping]
        )
        revert_run(out_files=out_files, dataset_name=self.dataset_name)

    def check_updates(self, n_latest: int = 9, check_o_url: bool = True):
        """Shows what inputs are supposed to be updated- and imported upon omni_obj.update_object()

        Args:
            n_latest (int, optional): Number of latest pipelines to check for the orchestrator check. Defaults to 9.
            check_o_url (bool, optional): If the imported datasets have to be part of an orchestrator. Defaults to True.
        """

        imp_list: List = []
        up_list: List = []

        if self.inputs is None:
            print(
                "No inputs defined and no input keyword provided.\n"
                "Nothing will be updated.\n"
                "Consider adding an instance of OmniInput as omni_obj.inputs.\n"
                "Imports will be based on omni_obj.inputs.keywords."
            )
            return
        if self.inputs.keyword is None:
            print(
                "No keyword specified. Nothing to update/import.\n"
                "Please provide a keyword to enable imports in the form of: \n"
                "omni_obj.inputs.keyword = [KEY1, KEY2]"
            )
            return
        if not check_o_url:
            self.orchestrator = "placeholder/string"
        if self.orchestrator is None:
            if self.benchmark_name is not None:
                self.orchestrator = find_orchestrator(
                    benchmark_name=self.benchmark_name, bench_url=self.BENCH_URL
                )
            if self.orchestrator is None:
                print(
                    f"WARNING: No orchestrator specified! \n"
                    f"No new datasets will be imported. Consider specifying an orchestrator by running:\n"
                    f"OmniObject.orchestrator = find_orchestrator(BENCHMARK_NAME) \n"
                    f"Look at {self.BENCH_URL} to get a list of possible BENCHMARK_NAMEs."
                )
                return
        if self.parameter is None:
            keys = self.inputs.keyword
        else:
            keys = flatten(
                [
                    k
                    for k in [self.inputs.keyword, self.parameter.keyword]
                    if k is not None
                ]
            )
        for key in keys:
            imp_ids, up_names = get_data_url_by_keyword(
                keyword=key,
                filter_names=self.inputs.filter_names,
                o_url=self.orchestrator,
                filter_ex=True,
                query_url=self.DATA_QUERY_URL,
                data_url=self.DATA_URL,
                gitlab_url=self.GIT_URL,
                check_o_url=check_o_url,
                n_latest=n_latest,
            )

            imp_list.append(imp_ids)
            up_list.append(up_names)

        imp_list = flatten(imp_list)
        up_list = flatten(up_list)
        nl = "\n"
        if len(imp_list) > 0:
            print(
                f"The following datasets will be imported upon omni_obj.update_object():\n"
                f"{nl}{nl.join(imp_list)}"
            )

        if len(up_list) > 0:
            print(
                f"The following datasets will be updated upon omni_obj.update_object():\n"
                f"{nl}{nl.join(up_list)}"
            )
        else:
            print(
                "No existing dataset matching the specified keyword found. Nothing to be updated"
            )

    def check_run(self):

        self.command = check_omni_command(self.command, self.script, self.outputs)
        out_files = get_all_output_file_names(self.outputs)
        plan = check_plan_exist(out_files)
        nl = "\n"
        if plan is None:
            print(
                "No workflow associated to this object exist.\n"
                "The following command will be run to generate one:\n"
                f"{self.command.command_line}\n"
                "Further outputs will be generated with that workflow:\n"
                f"{nl}{nl.join(out_files)}"
            )
            return
        plan_view = PlanViewModel.from_plan(plan)
        activity_out = filter_activity_exist(out_files)
        activity_out = rm_none_from_list(activity_out)
        no_activity = [out for out in out_files if out not in activity_out]
        print(
            "The following workflow is associated to this object:\n"
            f"Command:\n {plan_view.full_command}\n"
            f"Inputs:\n {[input.__dict__ for input in plan_view.inputs]}\n"
            f"Outputs:\n {[output.__dict__ for output in plan_view.outputs]}\n"
            f"Parameter:\n {[param.__dict__ for param in plan_view.parameters]}\n"
            "Activities for the following outputs exist and will be checked for updates:\n"
            f"{nl}{nl.join(activity_out)}\n"
            "New activities for the following outputs will  be generated:\n"
            f"{nl}{nl.join(no_activity)}"
        )

    def execute_plan_without_KG(
        self, inputs, parameter, outputs, time_out: Optional[int] = None
    ):
        pass



""" All test around the OmniParameter class and related functions"""

from omnibenchmark.core.input_classes import OmniParameter
from omnibenchmark.utils.exceptions import InputError
import pytest

### Test OmniParameter class


def test_omni_parameter_with_values(mock_param_values):
    omni_p = OmniParameter(names=["param1", "param2"], values=mock_param_values)
    assert isinstance(omni_p, OmniParameter)


def test_omni_parameter_default(mock_param_values):
    omni_p = OmniParameter(names=["param1", "param2"], values=mock_param_values)
    assert omni_p.default == {"param1": 0, "param2": "test"}


def test_omni_parameter_value_names(mock_param_values):
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniParameter(names=["no_param", "param2"], values=mock_param_values)


def test_omni_parameter_comb_names(mock_combinations):
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniParameter(names=["no_param", "param2"], combinations=mock_combinations)



"""Default settings - overwrite locally"""


class OmniRenkuInst(object):
    """Class to store the omnibenchmark renku configuration"""

    KG_URL = "https://renkulab.io/knowledge-graph"
    DATA_URL = KG_URL + "/datasets/"
    DATA_QUERY_URL = KG_URL + "/datasets?query="
    RENKU_URL = "https://renkulab.io"
    GIT_URL = "https://renkulab.io/gitlab"
    GRAPHQL_URL = "https://renkulab.io/knowledge-graph/graphql"
    GIT_API = "https://renkulab.io/gitlab/api/v4/"
    BENCH_URL = "https://omnibenchmark.pages.uzh.ch/omb-site/p/benchmarks"



# -*- coding: utf-8 -*-
#
# Copyright 2022-2022- University of Zuerich (UZH)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Omnibenchmark"""

# Version of the omnibenchmark package

__version__ = "0.0.41"


from omnibenchmark import core, management, renku_commands, utils
from omnibenchmark.core.omni_object import OmniObject
from omnibenchmark.core.output_classes import OmniOutput, OmniCommand, OmniPlan
from omnibenchmark.core.input_classes import OmniParameter, OmniInput
from omnibenchmark.utils.build_omni_object import get_omni_object_from_yaml
from omnibenchmark.renku_commands.general import renku_save
from omnibenchmark.management.data_commands import update_dataset_files
from omnibenchmark.management.data_commands import get_data_url_by_keyword

__all__ = (
    "core",
    "management",
    "renku_commands",
    "utils",
    "OmniObject",
    "OmniPlan",
    "OmniOutput",
    "OmniCommand",
    "OmniParameter",
    "OmniInput",
    "get_omni_object_from_yaml",
    "renku_save",
    "update_dataset_files",
    "get_data_url_by_keyword",
)



from omnibenchmark.core.input_classes import OmniInput, OmniParameter
from omnibenchmark.core.output_classes import OmniOutput
import pytest
import requests
from renku.domain_model.dataset import Dataset as RenkuDataset
from renku.domain_model.dataset import DatasetFile as RenkuDatasetFile
from renku.domain_model.entity import Entity
from renku.ui.api.models.dataset import Dataset as ApiDataset
from renku.ui.api.models.dataset import DatasetFile as ApiDatasetFile
from renku.command.view_model.plan import (
    PlanViewModel,
    CommandInputViewModel,
    CommandOutputViewModel,
    CommandParameterViewModel,
)
from renku.domain_model.workflow.plan import Plan
from renku.domain_model.workflow.parameter import (
    CommandOutput,
    CommandInput,
    CommandParameter,
)
from renku.domain_model.provenance.activity import Activity, Association
from renku.infrastructure.gateway.activity_gateway import ActivityGateway
from renku.ui.api.util import get_activity_gateway
from omnibenchmark.core.input_classes import OutMapping
import renku.ui.api
import omnibenchmark.management.general_checks
import omnibenchmark.renku_commands.renku_api


### API related fixtures


@pytest.fixture
def disable_api_calls(monkeypatch):
    def stunted_get():
        raise RuntimeError("Network access not allowed during testing!")

    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: stunted_get())


@pytest.fixture
def mock_html_content():
    return b'\n<!doctype html>\n<html>\n<body>\n<table>\n<thead>\n<tr>\n<th>Benchmark_name</th>\n<th>Description</th>\n<th>Orchestrator</th>\n</tr>\n</thead>\n<tbody>\n<tr>\n<td><em>omni_batch</em>, <em>batch_integration</em></td>\n<td>Methods to integrate/correct batch effects in sc-RNA-seq data</td>\n<td><a href="https://renkulab.io/knowledge-graph/projects/omnibenchmark/orchestrator">https://renkulab.io/knowledge-graph/projects/omnibenchmark/orchestrator</a></td>\n</tr>\n<tr>\n<td><em>omni_celltype</em>, ...</td>\n<td></td>\n<td></td>\n</tr>\n</tbody>\n</table>\n</body>\n</html>'


@pytest.fixture
def mock_dataset_json():
    return [
        {
            "_links": [{"rel": "details", "href": "https://this_is_a_mo.ck"}],
            "description": "A mock dataset that does not exoist in real",
            "identifier": "XXXXXX",
            "title": "A mock dataset",
            "name": "mock_dataset",
            "published": {
                "creator": [{"email": "mock.test@mail.com", "name": "Tesuto Mock"}]
            },
            "date": "2021-06-10T07:17:51.817557Z",
            "projectsCount": 0,
            "keywords": ["mock"],
            "images": [],
        }
    ]


@pytest.fixture
def mock_dataset_info():
    return {
        "_links": [
            {
                "rel": "self",
                "href": "https://renkulab.io/knowledge-graph/datasets/asdfghjjkl8765r",
            },
            {
                "rel": "initial-version",
                "href": "https://renkulab.io/knowledge-graph/datasets/asdfghjjkl8765r",
            },
        ],
        "identifier": "asdfghjjkl8765r",
        "name": "mock_dataset",
        "title": "mock dataset info",
        "url": "https://renkulab.io/datasets/asdfghjjkl8765r",
        "derivedFrom": "https://renkulab.io/datasets/asdfghjjkl8765r",
        "versions": {"initial": "assfsadfel8765r"},
        "description": "something to mock",
        "published": {
            "creator": [
                {"name": "Michael Mustermann", "email": "michi.mustermann@test.com"}
            ]
        },
        "created": "2022-05-25T13:54:51Z",
        "hasPart": [
            {"atLocation": "data/mock_dataset/counts_test.mtx.gz"},
            {"atLocation": "data/mock_dataset/data_info_test.json"},
        ],
        "project": {
            "_links": [
                {
                    "rel": "project-details",
                    "href": "https://renkulab.io/knowledge-graph/projects/mock/data_info",
                }
            ],
            "path": "mock/data_info",
            "name": "data_info",
        },
        "usedIn": [
            {
                "_links": [
                    {
                        "rel": "project-details",
                        "href": "https://renkulab.io/knowledge-graph/projects/mock/data_info",
                    }
                ],
                "path": "mock/data_info",
                "name": "data_info",
            }
        ],
        "keywords": ["mock"],
        "images": [],
    }


@pytest.fixture
def mock_response_json(mock_dataset_json):
    class MockResponse:
        @staticmethod
        def json():
            return mock_dataset_json

    return MockResponse()


@pytest.fixture
def mock_dataset_query(monkeypatch, mock_response_json):
    def get_mock_data_json():
        mock_res = mock_response_json
        return mock_res

    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_mock_data_json())


### Dataset related fixtures
@pytest.fixture
def mock_renkuDatasetFile():

    ent = Entity(path="some/path/to/genes_file.txt", checksum="XDFGHJK")
    mock_data_fi = RenkuDatasetFile(entity=ent)
    return mock_data_fi


@pytest.fixture
def mock_renkuApiDatasetFile(mock_renkuDatasetFile):
    mock_data_fi = ApiDatasetFile._from_dataset_file(mock_renkuDatasetFile)
    return mock_data_fi


@pytest.fixture
def mock_renkuDataset(mock_renkuDatasetFile):

    mock_data = RenkuDataset(name="mock_dataset")
    mock_data.keywords = ["mock"]
    mock_data.title = "This is a mock dataset and does not exist on the KG"
    mock_data.dataset_files = [mock_renkuDatasetFile]
    return mock_data


@pytest.fixture
def mock_renkuDataset_2files(mock_renkuDatasetFile):

    mock_data = RenkuDataset(name="mock_dataset")
    mock_data.keywords = ["mock"]
    mock_data.title = "This is a mock dataset and does not exist on the KG"
    ent = Entity(path="some/path/to/features_file.txt", checksum="XDFGHJKP")
    mock_data_fi = RenkuDatasetFile(entity=ent)
    mock_data.dataset_files = [mock_renkuDatasetFile, mock_data_fi]
    return mock_data


@pytest.fixture
def mock_api_Dataset(mock_renkuDataset):

    api_data = ApiDataset()
    api_data = api_data._from_dataset(mock_renkuDataset)
    return api_data


@pytest.fixture
def mock_api_Dataset_2files(mock_renkuDataset_2files):

    api_data = ApiDataset()
    api_data = api_data._from_dataset(mock_renkuDataset_2files)
    return api_data


@pytest.fixture
def get_renkuDataset_List(monkeypatch, mock_api_Dataset):
    def get_mock_list():
        return [mock_api_Dataset]

    # monkeypatch.setattr(
    #    renku.ui.api.models.dataset.Dataset,
    #    "list",
    #    lambda *args, **kwargs: get_mock_list(),
    # )

    monkeypatch.setattr(
        omnibenchmark.renku_commands.renku_api,
        "renku_dataset_list",
        lambda *args, **kwargs: get_mock_list(),
    )


@pytest.fixture
def no_project_context(monkeypatch):
    def no_renku_project_context():
        return False

    monkeypatch.setattr(
        omnibenchmark.management.general_checks,
        "is_renku_project",
        lambda *args, **kwargs: no_renku_project_context(),
    )


### Activity related fixtures


@pytest.fixture
def act_gateway():
    return get_activity_gateway()


@pytest.fixture
def mock_input_view():

    input = CommandInputViewModel(name="input1", default_value="this/is/a/default")
    return input


@pytest.fixture
def mock_output_view():

    output = CommandOutputViewModel(
        name="output1",
        default_value="this/is/another/default",
        create_folder=False,
    )
    return output


@pytest.fixture
def mock_omni_output_plan():
    out_map = OutMapping(
        {"output_files": {"output1": "this/is/another/default"}}
    )  # type:ignore
    omni_out = OmniOutput(
        name="mock_plan_out", out_names=["output1"], file_mapping=[out_map]
    )
    return omni_out


@pytest.fixture
def mock_param_view():

    param = CommandParameterViewModel(
        name="param1", default_value="this/is/yet/another/default"
    )
    return param


@pytest.fixture
def mock_plan_view(mock_param_view, mock_input_view, mock_output_view):

    plan = PlanViewModel(
        id="XXX",
        command="test command",
        full_command="test command",
        name="test",
        created="2022-01-01",
        keywords=[],
        inputs=[mock_input_view],
        outputs=[mock_output_view],
        parameters=[mock_param_view],
    )
    return plan


@pytest.fixture
def mock_plan():

    plan = Plan(id="XXX", command="test command")
    return plan


@pytest.fixture
def mock_activity(mock_plan):

    association = Association(id="YYY", plan=mock_plan, agent="Almut Luetge")
    act = Activity(
        id="YYY",
        association=association,
        ended_at_time="2013-05-24T23:59:59",
        started_at_time="2013-05-24T23:52:59",
        agents=["Almut Luetge"],
    )
    return act


@pytest.fixture
def mock_output():

    output = CommandOutput(id="ZZZ", default_value="test/output", name="test_output-1")
    return output


@pytest.fixture
def mock_input():

    input = CommandInput(id="AAA", default_value="test/input", name="test_input-1")
    return input


@pytest.fixture
def mock_parameter():

    param = CommandParameter(id="AAA", default_value="test/param", name="test_param-1")
    return param


### Output/Input related fixtures
@pytest.fixture
def mock_prefix():
    return {"count_file": "counts", "dim_red_file": ["features", "genes"]}


@pytest.fixture
def mock_inputfiles():
    return {
        "data1": {"dim_red_file": "path/to/dim/red", "count_file": "path/to/count"},
        "data2": {"dim_red_file": "path/to/dim/red2", "count_file": "path/to/count2"},
    }


@pytest.fixture
def mock_omni_input(mock_inputfiles):
    input_obj = OmniInput(
        names=["dim_red_file", "count_file"], input_files=mock_inputfiles
    )
    return input_obj


@pytest.fixture
def mock_param_values():
    return {"param1": [0, 10], "param2": ["test"]}


@pytest.fixture
def mock_combinations():
    return [{"param1": 0, "param2": "test"}, {"param1": 10, "param2": "test"}]


@pytest.fixture
def mock_omni_parameter(mock_param_values, mock_combinations):
    param_obj = OmniParameter(
        names=["param1", "param2"],
        values=mock_param_values,
        combinations=mock_combinations,
    )
    return param_obj


# Omni output related fixtures


@pytest.fixture
def mock_out_mapping():
    map: OutMapping = {
        "output_files": {"out_file1": "path/to/out1", "out_file2": "path/to/out2"},
        "input_files": {"in_file1": "path/to/in1"},
        "parameter": {"param1": "str_value", "param2": 100},
    }

    return map


@pytest.fixture
def mock_plan_dict():
    map = {
        "out_file1": "out-1",
        "out_file2": "out-2",
        "in_file1": "in-1",
        "param1": "param-1",
        "param2": "param-2",
    }

    return map


@pytest.fixture
def mock_omni_output(mock_out_mapping, mock_omni_parameter, mock_omni_input):
    mock_out_mapping["input_files"] = mock_omni_input.input_files[
        mock_omni_input.default
    ]
    mock_out_mapping["parameter"] = mock_omni_parameter.default
    omni_out = OmniOutput(
        name="mock_res",
        out_names=["out_file1", "out_file2"],
        file_mapping=[mock_out_mapping],
        inputs=mock_omni_input,
        parameter=mock_omni_parameter,
    )
    return omni_out


@pytest.fixture
def mock_file_list(mock_combinations, mock_inputfiles):
    param = mock_combinations[0]
    param["param3"] = 0
    return [mock_inputfiles["data1"], param]



"""General checks to ensure the enviroment and setup are as expected"""

from renku.ui.api.models.project import Project
from omnibenchmark.utils.exceptions import InputError
from typing import Union, Optional, List
import os
import lxml.html as lh
import requests


def is_renku_project(path: Union[os.PathLike, str] = os.getcwd()) -> bool:
    """Checks if a project is an initialized renku project.
    Args:
        path (Pathlike, str, optional): A path to check. Defaults to ".".

    Returns:
        bool: True if project is a renku project.
    """
    current_dir = os.getcwd()
    os.chdir(path)
    project = Project()
    repo = project.repository
    os.chdir(current_dir)
    return True if repo.contains(".renku/metadata") else False


def find_orchestrator(
    benchmark_name: str,
    bench_url: str = "https://omnibenchmark.pages.uzh.ch/omb-site/p/benchmarks",
    key_header: str = "Benchmark_name",
    o_header: str = "Orchestrator",
) -> Optional[str]:
    """Assigns an orchestrator url based on the benchmark name by checking the omnibenchmark website.

    Args:
        benchmark_name (str): Name of the benchmark that thge object is part of.
        bench_url (_type_, optional): url to the omnibenchmark site with benchmark associations.
        key_header (str, optional): Header/column name specifying the benchmark names. Defaults to "Benchmark_name".
        o_header (str, optional): Header/column name specifying the orchestrator names. Defaults to "Orchestrator".

    Raises:
        InputError: Key_header and o_header need to be valid column names

    Returns:
        Optional[str]: Url to the orchestrator taht is linked to a specified benchmark.
    """
    bench_html = requests.get(bench_url)
    doc = lh.fromstring(bench_html.content)
    tr_elements = doc.xpath("//tr")
    header = tr_elements[0]
    col_num = 0
    for field in header:
        if field.text_content() == key_header:
            nam_col = col_num
        if field.text_content() == o_header:
            o_col = col_num
        col_num += 1

    try:
        nam_col, o_col
    except Exception:
        raise InputError(
            f"Could not find columns with names {key_header} and/or {o_header}.\n"
            f"Please check {bench_url} for the correct column names."
        )

    o_obj = [
        tr_ele[o_col]
        for tr_ele in tr_elements
        if benchmark_name in tr_ele[nam_col].text_content().split(",")
    ]
    if len(o_obj) < 1:
        print(
            f"WARNING: Could not find benchmark associated to {benchmark_name}.\n"
            f"Check {bench_url} for existing benchmarks.\n"
            f"Integration with existing projects is not possible."
        )
        return None
    o_url = o_obj[0].text_content()
    return o_url


def get_benchmark_groups(
    field_name: str,
    bench_url: str = "https://omnibenchmark.pages.uzh.ch/omb-site/p/benchmarks",
) -> List[str]:
    """Get all available benchmark groups as listed on bench_url

    Args:
        field_name (str): Table header name, with the entries to display
        bench_url (_type_, optional): Defaults to "https://omnibenchmark.pages.uzh.ch/omb-site/p/benchmarks".

    Raises:
        InputError: Raised if field name can not be found at bench_url

    Returns:
        List[str]: List with all entries for field_name at bench_url
    """
    bench_html = requests.get(bench_url)
    doc = lh.fromstring(bench_html.content)
    tr_elements = doc.xpath("//tr")
    header = tr_elements[0]
    col_num = 0
    for field in header:
        if field.text_content() == field_name:
            field_col = col_num
        col_num += 1

    try:
        field_col
    except Exception:
        raise InputError(
            f"Could not find columns with names {field_name}.\n"
            f"Please check {bench_url} for the correct column names."
        )

    entries = [tr_ele[field_col].text_content() for tr_ele in tr_elements[1:]]
    return entries



"""Classes related to omnibenchmark outputs"""

from typing import Mapping, List, Optional, Union, Callable
from omnibenchmark.utils.user_input_checks import check_name_matching, flatten
from omnibenchmark.utils.exceptions import OutputError
from omnibenchmark.utils.auto_output import (
    get_all_output_combinations,
    get_default_outputs,
    autocomplete_file_mapping,
    filter_file_mapping_list,
)
from omnibenchmark.utils.auto_run import map_plan_names_file_types
from omnibenchmark.core.input_classes import OmniInput, OmniParameter, OutMapping
from omnibenchmark.utils.auto_command import (
    automatic_command_generation,
    get_interpreter_from_extension,
)
from renku.command.view_model.plan import PlanViewModel
from os import PathLike
import os


class OmniOutput:
    """Class to store metadata of datasets and files that are specified as output in an omnibenchmark project"""

    def __init__(
        self,
        name: str,
        out_names: List[str],
        output_end: Optional[Mapping[str, str]] = None,
        out_template: str = "data/${name}/${name}_${unique_values}_${out_name}.${out_end}",
        file_mapping: Optional[List[OutMapping]] = None,
        inputs: Optional[OmniInput] = None,
        parameter: Optional[OmniParameter] = None,
        default: Optional[Mapping] = None,
        filter_json: Optional[str] = None,
        sort_keys: bool = True,
        template_fun: Optional[Callable[..., Mapping]] = None,
        template_vars: Optional[Mapping] = None,
    ):
        """Class to manage outputs of an omnibenchmark module

        Args:
            name (str): Name that is specific for all outputs. Typically the omnibenchmark module name/OmniObject name
            out_names (List[str]): Names of the output file types
            output_end (Optional[Mapping[str, str]], optional): Endings of the output filetypes. Defaults to None.
            out_template (str, optional): Template to generate output file names.
            file_mapping (Optional[List[OutMapping]], optional): Mapping of input files, parameter values and output files.
                                                                 Defaults to None.
            inputs (Optional[OmniInput], optional): Object specifying all valid inputs. Defaults to None.
            parameter (Optional[OmniParameter], optional): Object speccifying the parameter space. Defaults to None.
            default (Optional[Mapping], optional): Default output files. Defaults to None.
            filter_json(Optional[str], optional): Path to json file with filter combinations. Defaults to None.
            sort_keys(bool): If parameter keys should be sorted alphabetically to generate output names. Defaults to True.
            template_fun (Optional[Callable[..., Mapping]], optional): Function to automatically generate output filenames.
            template_vars (Optional[Mapping], optional): Variables that are used by template_fun. Defaults to None.
        """
        self.name = name
        self.out_names = out_names
        self.output_end = output_end
        self.out_template = out_template
        self.file_mapping = file_mapping
        self.inputs = inputs
        self.parameter = parameter
        self.default = default
        self.filter_json = filter_json
        self.sort_keys = sort_keys
        self.template_fun = template_fun
        self.template_vars = template_vars if template_vars is not None else {}

        if self.file_mapping is None:

            if self.output_end is None:
                raise OutputError(
                    "Undefined outputs. Please specify an output file mapping \n"
                    "or output prefix and output ending to automatically generate one."
                )

            check_name_matching(self.out_names, self.output_end.keys())
            self.file_mapping = get_all_output_combinations(
                name=self.name,
                output_end=self.output_end,
                out_template=self.out_template,
                inputs=self.inputs,
                parameter=self.parameter,
                sort_keys=self.sort_keys,
                template_fun=self.template_fun,
                **self.template_vars,
            )

        self.file_mapping = autocomplete_file_mapping(self.file_mapping)
        self.file_mapping = filter_file_mapping_list(
            file_mapping_list=self.file_mapping,
            inputs=self.inputs,
            parameter=self.parameter,
            filter_json=self.filter_json,
        )

        if self.file_mapping is not None:
            check_name_matching(
                self.out_names,
                flatten(
                    [
                        out_mapping["output_files"].keys()  # type: ignore
                        for out_mapping in self.file_mapping
                    ]
                ),
            )

            if self.default is None:
                self.default = get_default_outputs(
                    file_mapping=self.file_mapping,
                    inputs=self.inputs,
                    parameter=self.parameter,
                )
        if self.default is not None:
            check_name_matching(self.out_names, self.default.keys())

    def update_outputs(self):
        """Update output definitions according to the specified inputs/parameter. Does not update workflows or activities."""
        if self.inputs is not None or self.parameter is not None:
            self.template_vars = (
                self.template_vars if self.template_vars is not None else {}
            )
            new_file_mapping = get_all_output_combinations(
                name=self.name,
                output_end=self.output_end,
                out_template=self.out_template,
                inputs=self.inputs,
                parameter=self.parameter,
                sort_keys=self.sort_keys,
                template_fun=self.template_fun,
                **self.template_vars,
            )
            ex_file_mapping = self.file_mapping if self.file_mapping is not None else []
            all_file_mappings = ex_file_mapping + new_file_mapping
            self.file_mapping = list({str(i): i for i in all_file_mappings}.values())
            self.file_mapping = filter_file_mapping_list(
                file_mapping_list=self.file_mapping,
                inputs=self.inputs,
                parameter=self.parameter,
                filter_json=self.filter_json,
            )

        if self.file_mapping is not None:
            check_name_matching(
                self.out_names,
                flatten(
                    [
                        out_mapping["output_files"].keys()  # type: ignore
                        for out_mapping in self.file_mapping
                    ]
                ),
            )

            omni_outfiles = [file_map["output_files"] for file_map in self.file_mapping]
            if self.default is None or self.default not in omni_outfiles:
                self.default = get_default_outputs(
                    file_mapping=self.file_mapping,
                    inputs=self.inputs,
                    parameter=self.parameter,
                )


class OmniCommand:
    """Class to store metadata and attributes of a command in an omnibenchmark project"""

    def __init__(
        self,
        script: Union[PathLike, str],
        interpreter: Optional[str] = None,
        command_line: Optional[str] = None,
        outputs: Optional[OmniOutput] = None,
        input_val: Optional[Mapping] = None,
        parameter_val: Optional[Mapping] = None,
    ):
        """Initialize an instance of the class OmniCommand

        Args:
            script (Union[PathLike, str]): Path to the script run by the command
            interpreter (str, optional): Interpreter to run the script with. Defaults to None.
            command_line (str, optional): Command line to be run. Defaults to None.
            outputs (OmniOutput, optional): Object specifying all outputs. Defaults to None.
            input_val (Optional[Mapping], optional): Input file tyoes and paths to run the command on. Defaults to None.
            parameter_val (Optional[Mapping], optional): Parameter names and values to run the command with. Defaults to None.
        """
        self.script = script
        self.interpreter = interpreter
        self.command_line = command_line
        self.outputs = outputs
        self.input_val = (input_val,)
        self.parameter_val = parameter_val

        if self.command_line is None:
            if self.outputs is None or self.outputs.file_mapping is None:
                print(
                    "WARNING: No outputs/output file mapping in the current project detected.\n"
                    "Run OmniObject.update_object() to update outputs, inputs and parameter.\n"
                    "Currently this object can not be used to generate a workflow."
                )
            else:
                output_val = self.outputs.default
                self.input_val = next(
                    (
                        out_dict["input_files"]  # type:ignore
                        for out_dict in self.outputs.file_mapping
                        if out_dict["output_files"] == output_val
                    ),
                    None,  # type:ignore
                )
                self.parameter_val = next(
                    (
                        out_dict["parameter"]
                        for out_dict in self.outputs.file_mapping
                        if out_dict["output_files"] == output_val
                    ),
                    None,
                )
                if self.interpreter is None:
                    ext = os.path.splitext(self.script)[1]
                    self.interpreter = get_interpreter_from_extension(ext)

                self.command_line = automatic_command_generation(
                    self.script,
                    interpreter=self.interpreter,
                    inputs=self.input_val,  # type:ignore
                    outputs=output_val,
                    parameters=self.parameter_val,
                )
        # Add command checks!

    def update_command(self):
        """Update command according to the specifed inputs/parameter/output

        Raises:
            InputError: Needs an OmniOutput object or the command line specified.
        """
        if self.outputs is None or self.outputs.file_mapping is None:
            print(
                "WARNING: No outputs/output file mapping in the current project detected.\n"
                "Run OmniObject.update_object() to update outputs, inputs and parameter.\n"
                "Currently this object can not be used to generate a workflow."
            )
        else:
            output_val = self.outputs.default
            self.input_val = next(
                (
                    out_dict["input_files"]  # type:ignore
                    for out_dict in self.outputs.file_mapping
                    if out_dict["output_files"] == output_val
                ),
                None,  # type:ignore
            )
            self.parameter_val = next(
                (
                    out_dict["parameter"]
                    for out_dict in self.outputs.file_mapping
                    if out_dict["output_files"] == output_val
                ),
                None,
            )
            if self.interpreter is None:
                ext = os.path.splitext(self.script)[1]
                self.interpreter = get_interpreter_from_extension(ext)
            self.command_line = automatic_command_generation(
                self.script,
                interpreter=self.interpreter,
                inputs=self.input_val,  # type:ignore
                outputs=output_val,
                parameters=self.parameter_val,
            )


class OmniPlan:
    """Class to store the workflow definition/plan of an Omniobject"""

    def __init__(
        self, plan: PlanViewModel, param_mapping: Optional[Mapping[str, str]] = None
    ):
        """Initialize an instance of the class OmniPlan

        Args:
            plan (PlanViewModel): A plan view model as defined in renku
            param_mapping (Optional[Mapping[str, str]], optional): A mapping between the component names of the plan
                                                                   and the OmniObject. Defaults to None.
        """
        self.plan = plan
        self.param_mapping = param_mapping

    def predict_mapping_from_file_dict(
        self, file_dict: Mapping[str, str]
    ) -> Mapping[str, str]:
        """Predict the corresponding mappings of input/parameter/output as defined in the plan and the OmniObject

        Args:
            file_dict (Mapping[str, str]): input/parameter/output mapping as stored in the OmniObject

        Returns:
            Mapping[str, str]: Mapping between the plan names of input/parameter/outputs and the OmniObject names.
        """
        return map_plan_names_file_types(plan=self.plan, file_dict=file_dict)



""" All test around the OmniInput class and related functions"""

from omnibenchmark.core.input_classes import OmniInput
from omnibenchmark.utils.exceptions import InputError
import pytest


### Test OmniInput class
def test_omni_input_with_input_files(mock_inputfiles):
    test_input = OmniInput(
        names=["dim_red_file", "count_file"], input_files=mock_inputfiles
    )
    assert isinstance(test_input, OmniInput)


def test_omni_input_with_input_files_not_match_names(mock_inputfiles):
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniInput(names=["no_name", "count_file"], input_files=mock_inputfiles)


def test_omni_input_no_input_files_no_prefix():
    with pytest.raises(InputError, match=r"Missing input file definition"):
        OmniInput(names=["dim_red_file", "count_file"])


def test_omni_input_no_input_files_with_prefix_not_match_names():
    with pytest.raises(InputError, match=r"Expected inputs from .*?"):
        OmniInput(
            names=["dim_red_file", "count_file"],
            prefix={"dim_red_file": ["some"], "no_name": ["another"]},
            keyword=["some"],
        )


def test_omni_input_no_default(mock_inputfiles):
    test_input = OmniInput(
        names=["dim_red_file", "count_file"], input_files=mock_inputfiles
    )
    assert test_input.default == "data1"


def test_omni_input_no_default_files(mock_inputfiles):
    test_input = OmniInput(
        names=["dim_red_file", "count_file"], input_files=mock_inputfiles
    )
    assert test_input.default_files == mock_inputfiles["data1"]


def test_omni_input_default_name_mismatch(mock_inputfiles):
    with pytest.raises(InputError, match=r"Input default data3 not found.*?"):
        OmniInput(
            names=["dim_red_file", "count_file"],
            input_files=mock_inputfiles,
            default="data3",
        )


# def test_omni_input_get_input_files(mock_inputfiles, monkeypatch):
#    def get_mock_infiles():
#        return mock_inputfiles
#
#
#    monkeypatch.setattr(
#        omnibenchmark.utils.auto_input,
#        "get_input_files_from_prefix",
#        lambda *args, **kwargs: get_mock_infiles(),
#    )
#
#    test_input = OmniInput(names=["dim_red_file", "count_file"], prefix={"dim_red_file": ["some"], "count_file": ["another"]})
#
#    assert isinstance(test_input, OmniInput)
